/*
 * QUEUE.C - Generic queue handler module.  Include QUEUE.H
 *
 * $Header:   C:/fe/shared/vcs/queue.c_v   1.4   04 Jun 1990 02:22:50   rji  $
 *
 * $Log:   C:/fe/shared/vcs/queue.c_v  $
 * 
 *    Rev 1.4   04 Jun 1990 02:22:50   rji
 * qdelete(), formerly a void function, now returns a completion code.
 * 
 *    Rev 1.3   03 Sep 1989 23:11:54   rji
 * Changed #include file specifiers from <...> to "..." for non-standard
 * headers.
 * 
 *    Rev 1.2   14 Aug 1989  0:43:56   rji
 * Added new function qdelete() (and updated test driver accordingly.)
 * 
 *    Rev 1.1   09 Aug 1989 17:10:06   rji
 * Function headers changed to new ANSI format.
 * 
 *    Rev 1.0   07 Dec 1988 23:10:56   rji
 * Initial revision.
 * 
*/


/*=========================================================================*/
/*=                                                                       =*/
/*=      I N C L U D E S,  D E F I N E S,  D E C L A R A T I O N S        =*/
/*=                                                                       =*/
/*=========================================================================*/
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         QUEUE.C Includes

  $PATHS$
       Functions\C
       Shared\Includes
       Shared\Utils
       Modules\$./FL$

   $3$

   See Also:
        $/SEE(COMMON.H Macros)$
        $/SEE(DALLOC.H Macros)$
        $/SEE(QUEUE.H Macros)$

   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* INCLUDES                                                                */
/*-------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include "common.h"
#include "dalloc.h"

#include "queue.h"
/*                                                                   $end$ */

#if !defined(NDEBUG)
#define     queue_check(q,s)        qcheck((q),(s))
static void qcheck( QUEUE, char * );
#else
#define     queue_check(q,s)
#endif

/*-------------------------------------------------------------------------*/
/* PRIVATE FUNCTIONS                                                       */
/*-------------------------------------------------------------------------*/

static void qerror( char *, ... );


/*=========================================================================*/
/*=                                                                       =*/
/*=                F U N C T I O N   D E F I N I T I O N S                =*/
/*=                                                                       =*/
/*=========================================================================*/

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         qadd()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* QADD                                                                    */
/*                                                                         */
/* Adds element at rear of queue.  Caller supplies size of storage         */
/* required to hold queue element, and QADD allocates that amount plus     */
/* whatever storage is needed to maintain queue structure.  QADD returns   */
/* a pointer to an area the caller can then fill in for the particular     */
/* application purpose at hand.                                            */
/*-------------------------------------------------------------------------*/

void  *qadd(
   QUEUE    q,       /* in: queue to which element is to be added  */
   size_t   es       /* in: size of element to be added to queue   */
)
/*                                                                   $end$ */
{
   QUEUE_ELE_TOP *p;

   queue_check( q, "qadd" );    /* valid queue? */

   /* allocate requested space + queue management overhead space */

   if ( (p = (QUEUE_ELE_TOP *) d_malloc(es + sizeof(QUEUE_ELE_TOP))) == NULL )
      qerror( "qadd: no space for queue element\n" );

   /* link allocated space into queue */

   if ( q->front == NULL )
      q->front = p;
   else
      (q->rear)->next   = p;        /* previous rear element chains to new */

   q->rear = p;
   p->next = NULL;                  /* rear element has NULL next pointer  */

   ++(q->tally);                    /* increment queue count */

   return ( (void *) (p + 1) );     /* return pointer to user area */

} /* qadd() */


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         qcheck()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
#if !defined(NDEBUG)
/*-------------------------------------------------------------------------*/
/* QCHECK                                                                  */
/*                                                                         */
/* Checks integrity of a queue; can be turned off by re-compiling with     */
/* NDEBUG defined on compiler command line.                                */
/*-------------------------------------------------------------------------*/

static void qcheck(
   QUEUE    q,          /* in: queue whose integrity is to be checked      */
   char    *s           /* in: string to display if integrity check fails  */
)
/*                                                                   $end$ */
{
   if (   ( q == NULL )
       || ( q->front == NULL  &&  q->rear != NULL )
       || ( q->front != NULL  &&  q->rear == NULL )
       || ( q->front == NULL  &&  queue_length( q ) != 0 )
       || ( q->front != NULL  &&  queue_length( q ) == 0 )
       || ( queue_length( q ) < 0 )  )
      qerror( "qcheck: messed up queue (%s)\n", s );

   return;

} /* qcheck() */
#endif


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         qclear()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* QCLEAR                                                                  */
/*                                                                         */
/* Removes all elements from a queue.                                      */
/*-------------------------------------------------------------------------*/

void   qclear(
   QUEUE q           /* in: queue to be cleared */
)
/*                                                                   $end$ */
{
   QUEUE_ELE_TOP *p, *r;

   queue_check( q, "qclear" );    /* valid queue? */

   /* free all queue elements */

   for ( p = q->front;  p != NULL; )
   {
      r = p;
      p = p->next;
      d_free( r );
   }

   /* reset queue control variables */

   q->front = q->rear = NULL;
   q->tally = 0;

   return;

} /* qclear() */


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         qcreate()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* QCREATE                                                                 */
/*                                                                         */
/* Creates and returns a queue (allocates and initializes a queue header.) */
/*-------------------------------------------------------------------------*/

QUEUE qcreate()
/*                                                                   $end$ */
{
   QUEUE    q;

   if ( (q = (QUEUE) d_malloc( sizeof( QUEUE_HDR ) )) == NULL )
      qerror( "qcreate: no space for queue\n" );

   q->front = NULL;
   q->rear  = NULL;
   q->tally = 0;

   return ( q );

} /* qcreate() */


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         qdelete()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* QDELETE                                                                 */
/*                                                                         */
/* Deletes a queue element (frees element storage & adjusts predecessor's  */
/* next pointer).                                                          */
/* Returns:                                                                */
/*             0 = element could not be deleted                            */
/*             1 = element successfully deleted                            */
/*-------------------------------------------------------------------------*/

int   qdelete(
   QUEUE    q,    /* in/out: queue from which element is to be deleted     */
   void   **epp   /* in/out: pointer to pointer to queue element to delete */
)
/*                                                                   $end$ */
{
   QUEUE_ELE_TOP *p;

   queue_check( q, "qdelete" );        /* valid queue? */

   /* if pointer to element pointer is invalid, return immediately */

   if ( epp == NULL  ||  *epp == NULL )
      return ( 0 );

   /* loop through queue, looking for queue element to delete */

   for ( p = (QUEUE_ELE_TOP *) &q->front;  p->next != NULL;  p = p->next )
      if ( ((p->next) + 1) == *epp )
         break;

   /* if element found, free element's storage and adjust queue */

   if ( p->next != NULL )
   {
      p->next = ((QUEUE_ELE_TOP *) *epp - 1)->next;   /* fix pred next ptr */
      if ( p->next == NULL )     /* if rear to be deleted, adjust rear ptr */
         q->rear = (p == (QUEUE_ELE_TOP *) &q->front) ? NULL : p;
      d_free( ((QUEUE_ELE_TOP *) *epp - 1) );      /* free element storage */
      *epp = NULL;                     /* prevent access to freed storage  */
      --(q->tally);                             /* decrement queue count   */
      return ( 1 );
   }
   else
      return ( 0 );

} /* qdelete() */


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         qdestroy()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* QDESTROY                                                                */
/*                                                                         */
/* Destroys a queue by freeing all storage allocated to it.                */
/*-------------------------------------------------------------------------*/

void  qdestroy(
   QUEUE   *qp    /* in/out: pointer to queue (i.e., to queue header ptr)  */
)
/*                                                                   $end$ */
{
   QUEUE_ELE_TOP *p, *r;

   queue_check( *qp, "qdestroy" );     /* valid queue? */

   /* free all queue elements */

   for ( p = (*qp)->front;  p != NULL; )
   {
      r = p;
      p = p->next;
      d_free( r );
   }

   d_free( *qp );                      /* free queue header */
   *qp = NULL;                         /* prevent access to freed storage */

   return;

} /* qdestroy() */


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         qfind()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* QFIND                                                                   */
/*                                                                         */
/* Search for queue element & return pointer to it, if found.              */
/*-------------------------------------------------------------------------*/

void    *qfind(
   QUEUE    q,       /* in: queue in which element is to be found          */
   void    *key,     /* in: search key to be compared with queue elements  */
   int      (*cmpfn)( void *, void * )    /* in: function that compares key
                                          **     with queue element; returns
                                          **     zero iff key matches queue
                                          **     element.
                                          */
)
/*                                                                   $end$ */
{
   QUEUE_ELE_TOP *p;

   queue_check( q, "qfind" );    /* valid queue? */

   /* search queue elements for key match */

   for ( p = q->front;  p != NULL;  p = p->next )
      if ( !(*cmpfn)( key, (void *)(p + 1) ) )
         return ( (void *) (p + 1) );

   /* if this point reached, no match found */

   return ( NULL );

} /* qfind() */


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         qinsert()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* QINSERT                                                                 */
/*                                                                         */
/* Inserts element into queue in key order (set by a comparison function). */
/* Caller supplies key and size of storage required to hold queue element; */
/* QINSERT allocates that amount plus storage needed to maintain queue     */
/* structure.  Returns a pointer to an area the caller can then fill in    */
/* for the particular application purpose at hand.                         */
/*-------------------------------------------------------------------------*/

void    *qinsert(
   QUEUE    q,          /* in: queue in which element is to be inserted    */
   size_t   es,         /* in: size of element to be added to queue        */
   void    *key,        /* in: key of element to be inserted into queue    */
   int      (*cmpfn)( void *, void * )    /* in: function that compares key
                                          **     with queue element; returns
                                          **     <= 0 iff key s/b inserted
                                          **     before queue element.
                                          */
)
/*                                                                   $end$ */
{
   QUEUE_ELE_TOP *p, *r;

   queue_check( q, "qinsert" );        /* valid queue? */

   /* allocate requested space + queue management overhead space */

   if ( (r = (QUEUE_ELE_TOP *) d_malloc(es + sizeof(QUEUE_ELE_TOP))) == NULL )
      qerror( "qinsert: no space for queue element\n" );

   ++(q->tally);                                /* increment queue count   */

   /* find insertion point and link in storage */

   for ( p = (QUEUE_ELE_TOP *) &q->front;  p->next != NULL;  p = p->next )
      if ( (*cmpfn)( key, (void *)(p->next + 1) ) <= 0 )
      {
         r->next = p->next;
         p->next = r;
         return ( (void *) (r + 1) );
      }

   /* this point reached if queue empty or all elements failed comparison  */

   r->next = p->next;
   p->next = r;                  /* this sets q->front, if queue is empty  */
   q->rear = r;                  /* new element must be rear element       */

   return ( (void *) (r + 1) );            /* return pointer to user area  */

} /* qinsert() */


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
        qitcreat()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* QITCREAT                                                                */
/*                                                                         */
/* Creates a queue iterator to be used for passing through a queue from    */
/* the front to the rear.  Returns the queue iterator (i.e., a pointer to  */
/* a queue iterator structure).  See QITNEXT/QITDSTRY.                     */
/*-------------------------------------------------------------------------*/

QUEUE_ITERATOR qitcreat(
   QUEUE q
)
/*                                                                   $end$ */
{
   QUEUE_ITERATOR qi;

   if ( (qi = d_malloc( sizeof(QUEUE_ITERATOR_STRUCT) )) == NULL )
      qerror( "qitcreat: no space for queue\n" );

   qi->q = q;                             /* save queue reference          */
   qi->p = (QUEUE_ELE_TOP *) &q->front;   /* points to queue front pointer
                                          ** so first call to qitnext()
                                          ** returns pointer to first element
                                          */

   return( qi );

} /* qitcreat() */


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         qitdstry()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* QITDSTRY                                                                */
/*                                                                         */
/* Destroys a queue iterator by freeing storage allocated to it.           */
/* See QITCREAT/QITNEXT.                                                   */
/*-------------------------------------------------------------------------*/

void  qitdstry(
   QUEUE_ITERATOR   *qip
)
/*                                                                   $end$ */
{
   d_free( *qip );                     /* free QUEUE_ITERATOR_STRUCT       */
   *qip = NULL;                        /* prevent access to freed storage  */

} /* qitdstry() */


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         qitnext()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* QITNEXT                                                                 */
/*                                                                         */
/* Returns pointer to the next (first) element in queue for which the      */
/* specified iterator was created.  Returns NULL if queue is empty or      */
/* there is no next element.  See QITCREAT/QITDSTRY.                       */
/*-------------------------------------------------------------------------*/

void *qitnext(
   QUEUE_ITERATOR qi
)
/*                                                                   $end$ */
{
   if ( (qi->p->next) != NULL )
   {
      qi->p = qi->p->next;
      return( (void *) (qi->p + 1) );
   }

   return( NULL );

} /* qitnext() */


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         qerror()

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*-------------------------------------------------------------------------*/
/* QERROR                                                                  */
/*                                                                         */
/* Reports queue handling errors.                                          */
/*-------------------------------------------------------------------------*/

static void qerror( char *fmt, ... )
/*                                                                   $end$ */
{
   va_list  ap;

   /* can add qerror prefix message here, if desired */

   va_start( ap, fmt );
   vfprintf( stderr, fmt, ap );
   va_end( ap );

   exit( 1 );

} /* qerror() */


#if defined(DRIVER)

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
   $NAME$:
         QUEUE.C main() Test Driver

  $PATHS$
       Functions\C
       Shared\Procedures
       Shared\Utils
       Modules\$./FL$


   $1$

  $0$
  SYNOPSIS:
  $-1$
*/
/*=========================================================================*/
/*=                                                                       =*/
/*=                       T E S T   D R I V E R                           =*/
/*=                                                                       =*/
/*=========================================================================*/

#include <string.h>

void  main( void );
int   qcomp( void *, void * );

void  main()
/*                                                                   $end$ */
{
   enum optypes
   {
      QUIT, ADD, DELETE, INSERT, FIND, FIRST, LAST, LENGTH, ISEMPTY, CLEAR,
      FOR_DUMP, NEXT_DUMP
   } op;
   QUEUE             q;
   QUEUE_ITERATOR    qi;
   char              p[81], *pe;

   if ( (q = queue_create()) == NULL )
      fprintf( stderr, "queue_create failure\n" );

   for (;;)    /* manipulate queue loop */
   {
      printf( "\nEnter operation [and string, for ADD/DELETE/INSERT/FIND] "
              "where\n\n" );
      printf( "%d = QUIT\n%d = ADD\n%d = DELETE\n%d = INSERT\n%d = FIND\n"
              "%d = FIRST\n%d = LAST\n%d = LENGTH\n%d = ISEMPTY\n%d = CLEAR\n"
              "%d = FOR_DUMP\n%d = NEXT_DUMP"
              "\n\n==> ", QUIT, ADD, DELETE, INSERT, FIND, FIRST, LAST,
                          LENGTH, ISEMPTY, CLEAR, FOR_DUMP, NEXT_DUMP );
      fflush( stdout );
      scanf( "%d", &op );
      if ( op == QUIT )
         break;

      switch ( op )
      {
         case ADD:
            scanf( "%s", p );
            if ( (pe = queue_add( q, strlen(p) + 1 )) == NULL )
               fprintf( stderr, "queue_add failure\n\n" );
            else
               strcpy( pe, p );
            break;
         case DELETE:
            scanf( "%s", p );
            if ( (pe = queue_find( q, p, qcomp )) == NULL )
               printf( "\"%s\" not found\n\n", p );
            queue_delete( q, pe );
            break;
         case INSERT:
            scanf( "%s", p );
            if ( (pe = queue_insert( q, strlen(p) + 1, p, qcomp )) == NULL )
               fprintf( stderr, "queue_insert failure\n\n" );
            else
               strcpy( pe, p );
            break;
         case FIND:
            scanf( "%s", p );
            if ( (pe = queue_find( q, p, qcomp )) == NULL )
               printf( "\"%s\" not found\n\n", p );
            else
               printf( "\"%s\" found (should = \"%s\")\n\n", p, pe );
            break;
         case FIRST:
            if ( !queue_isempty(q) )
               if ( (pe = queue_first( q )) == NULL )
                  fprintf( stderr, "queue_first failure\n\n" );
               else
                  printf( "First Element = \"%s\"\n\n", pe );
            else
               printf( "queue is empty\n\n" );
            break;
         case LAST:
            if ( !queue_isempty(q) )
               if ( (pe = queue_last( q )) == NULL )
                  fprintf( stderr, "queue_last failure\n\n" );
               else
                  printf( "Last Element = \"%s\"\n\n", pe );
            else
               printf( "queue is empty\n\n" );
            break;
         case LENGTH:
            printf( "%d\n\n", queue_length( q ) );
            break;
         case ISEMPTY:
            printf( "%s\n\n", queue_isempty( q ) ? "true" : "false" );
            break;
         case CLEAR:
            queue_clear( q );
            printf( "queue length = %d\n\n", queue_length( q ) );
            break;
         case FOR_DUMP:
            FOR_EACH_Q_ELE( q, pe, char * )
               printf( "\"%s\"\n", pe );
            printf( "\n" );
            break;
         case NEXT_DUMP:
            qi = queue_iterator_create( q );
            while ( (pe = queue_iterator_next( qi )) != NULL )
               printf( "\"%s\"\n", pe );
            printf( "\n" );
            queue_iterator_destroy( qi );
            break;
      } /* switch */

   } /* for */

   queue_destroy( q );

   exit( 0 );

} /* main() */

/*-------------------------------------------------------------------------*/

int   qcomp( key, ele )
   void *key;
   void *ele;
{
   return ( strcmp( (char *) key, (char *) ele ) );

} /* qcomp() */

#endif
